# frozen_string_literal: true

require_relative 'complex_type'
require_relative 'edm/primitive_types'
require_relative 'transition'

module Safrano
  def self.FunctionImport(name)
    FunctionImport::Function.new(name)
  end

  module FunctionImport
    class Function
      @allowed_transitions = [Safrano::TransitionEnd]
      attr_reader :name
      attr_reader :proc

      def initialize(name)
        @name = name
        @http_method = 'GET'
      end

      def allowed_transitions
        [Safrano::TransitionExecuteFunc]
      end

      def input(**parmtypes)
        @input = {}
        parmtypes.each do |k, t|
          @input[k] = case t.name
                      when 'Integer'
                        Safrano::Edm::Edm::Int32
                      when 'String'
                        Safrano::Edm::Edm::String
                      when 'Float'
                        Safrano::Edm::Edm::Double
                      when 'DateTime'
                        Safrano::Edm::Edm::DateTime
                      else
                        t
                      end
        end
        self
      end

      def auto_query_parameters
        @auto_query_params = true
        self # chaining
      end
      alias auto_query_params auto_query_parameters

      def return(klassmod, &proc)
        raise('Please provide a code block') unless block_given?

        @returning = if klassmod.respond_to? :return_as_instance_descriptor
                       klassmod.return_as_instance_descriptor
                     else
                       # if it's neither a ComplexType nor a Model-Entity
                       # --> assume it is a Primitive
                       ResultDefinition.asPrimitiveType(klassmod)
                     end
        @proc = proc
        self
      end

      def return_collection(klassmod, &proc)
        raise('Please provide a code block') unless block_given?

        @returning = if klassmod.respond_to? :return_as_collection_descriptor
                       klassmod.return_as_collection_descriptor
                     else
                       # if it's neither a ComplexType nor a Modle-Entity
                       # --> assume it is a Primitive
                       #                       ResultAsPrimitiveTypeColl.new(klassmod)
                       ResultDefinition.asPrimitiveTypeColl(klassmod)
                     end
        @proc = proc
        self
      end
      #      def initialize_params
      #        @uparms = UrlParameters4Func.new(@model, @params)
      #      end

      def check_missing_params
        # do we have all parameters provided ?  use Set difference to check
        pkeys = @params.keys.map(&:to_sym).to_set
        unless (idiff = @input.keys.to_set - pkeys).empty?

          Safrano::ServiceOperationParameterMissing.new(
            missing: idiff.to_a,
            sopname: @name
          )
        else
          Contract::OK
        end
      end

      def check_url_func_params
        @funcparams = {}
        return nil unless @input # anything to check ?

        # do we have all parameters provided ?
        check_missing_params.tap_error { |error| return error }
        # ==> all params were provided

        # now we shall check the content and type of the parameters
        @input.each do |ksym, typ|
          typ.convert_from_urlparam(v = @params[ksym.to_s])
             .tap_valid do |retval|
            @funcparams[ksym] = retval
          end
             .tap_error do
            # return is really needed here, or we end up returning nil below
            return parameter_convertion_error(ksym, typ, v)
          end
        end
        nil
      end

      def parameter_convertion_error(param, type, val)
        Safrano::ServiceOperationParameterError.new(type: type,
                                                    value: val,
                                                    param: param,
                                                    sopname: @name)
      end

      def add_metadata_rexml(ec)
        ## https://services.odata.org/V2/OData/Safrano.svc/$metadata
        # <FunctionImport Name="GetProductsByRating" EntitySet="Products" ReturnType="Collection(ODataDemo.Product)" m:HttpMethod="GET">
        # <Parameter Name="rating" Type="Edm.Int32" Mode="In"/>
        # </FunctionImport>
        funky = ec.add_element('FunctionImport',
                               'Name' => @name.to_s,
                               #                       EntitySet= @entity_set ,
                               'ReturnType' => @returning.type_metadata,
                               'm:HttpMethod' => @http_method)
        @input.each do |iname, type|
          funky.add_element('Parameter',
                            'Name' => iname.to_s,
                            'Type' => type.type_name,
                            'Mode' => 'In')
        end if @input
        funky
      end

      def with_transition_validated(req)
        #        initialize_params
        @params = req.params
        return yield unless (@error = check_url_func_params)

        [nil, :error, @error] if @error
      end

      def do_execute_func(req)
        with_transition_validated(req) do
          result = @proc.call(**@funcparams)
          [@returning.do_execute_func_result(result, req, @auto_query_params), :run]
        end
      end

      def transition_execute_func(_match_result)
        [self, :run_with_execute_func]
      end
    end
  end
end
