# frozen_string_literal: true

require 'rack'
require_relative '../odata/walker'
require_relative 'request'
require_relative 'response'

module Safrano
  # handle GET PUT etc
  module MethodHandlers
    def odata_options
      @walker.finalize.tap_error { |err| return err.odata_get(@request) }
             .if_valid do |context|
        # cf. stackoverflow.com/questions/22924678/sinatra-delete-response-headers
        headers.delete('Content-Type')
        @response.headers.delete('Content-Type')
        @response.headers['Content-Type'] = ''
        [200, EMPTY_HASH, '']
      end
    end

    def odata_delete
      @walker.finalize.tap_error { |err| return err.odata_get(@request) }
             .if_valid { |context| context.odata_delete(@request) }
    end

    def odata_put
      @walker.finalize.tap_error { |err| return err.odata_get(@request) }
             .if_valid { |context| context.odata_put(@request) }
    end

    def odata_patch
      @walker.finalize.tap_error { |err| return err.odata_get(@request) }
             .if_valid { |context| context.odata_patch(@request) }
    end

    def odata_get
      @walker.finalize.tap_error { |err| return err.odata_get(@request) }
             .if_valid { |context| context.odata_get(@request) }
    end

    def odata_post
      @walker.finalize.tap_error { |err| return err.odata_get(@request) }
             .if_valid { |context| context.odata_post(@request) }
    end

    def odata_head
      [200, EMPTY_HASH, [EMPTY_STRING]]
    end
  end

  # the main Rack server app. Source: the Rack docu/examples and partly
  # inspired from Sinatra
  class ServerApp
    METHODS_REGEXP = Regexp.new('HEAD|OPTIONS|GET|POST|PATCH|MERGE|PUT|DELETE').freeze
    NOCACHE_HDRS = { 'Cache-Control' => 'no-cache',
                     'Expires' => '-1',
                     'Pragma' => 'no-cache' }.freeze
    DATASERVICEVERSION = 'DataServiceVersion'
    include MethodHandlers

    def before
      @request.service_base = self.class.get_service_base

      @request.negotiate_service_version.tap_valid do
        myhdrs = NOCACHE_HDRS.dup
        myhdrs[DATASERVICEVERSION] = @request.service.data_service_version
        headers myhdrs
      end
    end

    # dispatch for all methods requiring parsing of the path
    # with walker (ie. allmost all excepted HEAD)
    def dispatch_with_walker
      @walker = @request.create_odata_walker
      case @request.request_method
      when 'GET'
        odata_get
      when 'POST'
        odata_post
      when 'DELETE'
        odata_delete
      when 'OPTIONS'
        odata_options
      when 'PUT'
        odata_put
      when 'PATCH', 'MERGE'
        odata_patch
      else
        raise Error
      end
    end

    def dispatch_error(err)
      @response.status, rsph, @response.body = err.odata_get(@request)
      headers rsph
    end

    def dispatch
      req_ret = if @request.request_method !~ METHODS_REGEXP
                  [404, EMPTY_HASH, ['Did you get lost?']]
                elsif @request.request_method == 'HEAD'
                  odata_head
                else
                  dispatch_with_walker
                end
      @response.status, rsph, @response.body = req_ret
      headers rsph
    end

    def call(env)
      # for thread safety
      dup._call(env)
    end

    def _call(env)
      begin
        @request = Safrano::Request.new(env)
        @response = Safrano::Response.new

        before.tap_error { |err| dispatch_error(err) }
              .tap_valid { |res| dispatch }

      # handle remaining Sequel errors that we couldnt prevent with our
      # own pre-checks
      rescue Sequel::Error => e
        dispatch_error(SequelExceptionError.new(e))
      end
      @response.finish
    end

    # Set multiple response headers with Hash.
    def headers(hash = nil)
      @response.headers.merge! hash if hash
      @response.headers
    end

    def self.enable_batch
      @service_base.enable_batch
    end

    def self.path_prefix(path_pr)
      @service_base.path_prefix path_pr
    end

    def self.get_service_base
      @service_base
    end

    def self.set_servicebase(sbase)
      @service_base = sbase
      @service_base.enable_v1_service
      @service_base.enable_v2_service
    end

    def self.publish_service(&block)
      sbase = Safrano::ServiceBase.new
      sbase.instance_eval(&block) if block_given?
      sbase.finalize_publishing
      set_servicebase(sbase)
    end
  end
end

# deprecated
# REMOVE 0.6
module OData
  class ServerApp < Safrano::ServerApp
    def self.publish_service(&block)
      ::Safrano::Deprecation.deprecate('OData::ServerApp',
                                       'Use Safrano::ServerApp instead')
      super
    end
  end
end
